/* -*- Mode: C++ -*- */

/* MemPortugal.C
 *
 * FCP Basic Agent
 * Luis Paulo Reis, Nuno Lau 2000 - Very simple version for demonstration
 * Copyright (C) 2000 Luis Paulo Reis, Nuno Lau
 *
 * Contacts:
 * Luis Paulo Reis 
 *    University Fernando Pessoa - Porto, Portugal
 *    Http://www.cerem.ufp.pt/~lpreis
 *    Email: lpreis@ufp.pt
 *    LIACC - Artificial Intelligence and Computer Science Lab.- University of Porto
 *    Http://www.fe.up.pt/~lpreis
 *    Email: lpreis@fe.up.pt
 * Nuno Lau
 *    University of Aveiro - Aveiro, Portugal
 *    Http://www.ua.pt/~lau
 *    Email: lau@det.ua.pt
 *
 * This Team is based on CMUnited99 source code from Peter Stone, Patrick Riley and 
 * Manuela Veloso. If you make any changes to CMUnited99 code, or have any comments 
 * about it, please send a message to Peter Stone (pstone@cs.cmu.edu).
 *
 * You may copy and distribute this code freely as long as you retain this notice.
 * If you think this simple agent was useful to you for creating a good team or 
 * if you want to discuss some RoboCup related issues, please send me an Email 
 * message (lpreis@fe.up.pt).
 *
 * We tried to keep everything simple, using the minimum amount of code. We also 
 * tried to create a competitve team and give ideas about how to transform it into 
 * a very good team!
 * 
 * This Team is able to play a decent game against all RoboCup99 teams (including 
 * CMUnited99) and RoboCup2000 teams (including FC Portugal). If you want to transform 
 * this simple team into a really competitive team, please read the suggestion 
 * included in the code and in the paper: [Reis and Lau 2001] Reis, Luis Paulo and 
 * Lau, Nuno, "FC Portugal Team Description: RoboCup 2000 Simulation League Champion", 
 * RoboCup 2000: Robot Soccer World Cup IV, Springer Verlag, 2001.
 *
 * For more information about the FC Portugal Project (European and World Champions of 
 * RoboCup 2000), please see Http://www.ieeta.pt/robocup or contact us.
 */

#include "MemPortugal.h"
#include "kick.h"
#include "mybehave.h"
#include "client.h"

void Portugal1Info::Initialize()
{
  ReadFormations();
  SetGlobalFormation(0);
  for(int i=1;i<=SP_team_size;i++) FM_Positioning[i]=i;
}

/****************************************************************************************/
/********************************* Formation Analysis ***********************************/
/****************************************************************************************/

// Why not use tactics composed by formations and player types? See [Reis and Lau 2001]
void Portugal1Info::ReadFormations()
{
  int i,t,k;
  FILE *fich;
  char line1[512], *str;

  if ((fich=fopen("formations.conf","r"))==NULL) { 
	fprintf(stderr,"Formation Configuration File does not Exist!\n"); return; }
  
  for(i=0; i < NUM_FORM; i++) {
        fgets(line1, 200, fich); fgets(line1, 30, fich);
        fgets(line1, 200, fich); str = &line1[0];
	for(t=1; t<=11; t++) FM_Position[i][0][t].x = get_float(&str); 
        fgets(line1, 200, fich); str = &line1[0];
	for(t=1; t<=11; t++) FM_Position[i][0][t].y = get_float(&str); 
        fgets(line1, 200, fich); str = &line1[0];
        for(t=1; t<=11; t++) FM_PlayerType[i][t] = get_int(&str); 
        fgets(line1, 200, fich); str = &line1[0];
        for(k=0; k<NUM_PLTYPES; k++) FM_Attraction[i][k].x = get_float(&str); 
        fgets(line1, 200, fich); str = &line1[0];
        for(k=0; k<NUM_PLTYPES; k++) FM_Attraction[i][k].y = get_float(&str); 
        fgets(line1, 200, fich); str = &line1[0];
        for(k=0; k<NUM_PLTYPES; k++) FM_BehindBall[i][k] = get_int(&str); 
        fgets(line1, 200, fich); str = &line1[0];
        for(k=0; k<NUM_PLTYPES; k++) FM_MinX[i][k] = get_float(&str); 
        fgets(line1, 200, fich); str = &line1[0];
        for(k=0; k<NUM_PLTYPES; k++) FM_MaxX[i][k] = get_float(&str); 
  }
  fclose(fich);
}

// Why not use situations? (see [Reis and Lau 2001]) What should be the tactic and formation 
// for each situation? How to define different player types?
void Portugal1Info::DecideFormation(void)
{
    SetGlobalFormation(0);
}

/****************************************************************************************/
/*********************************  Communication  **************************************/
/****************************************************************************************/

// Only players that think they have important things to say, should talk! 
void Portugal1Info::DoCommunication(void)   // Communicate World State
{
    if (ClockStopped || BallVelocityValid()>=0.9 && (CurrentTime.t+2) % 2 == 0 ) 
	send_game_status();  // Send World State
}

void Portugal1Info::say_something(char *message)   // Say Message
{
  SayAction.type = CMD_say;
  SayAction.time = CurrentTime;
  sprintf(SayAction.command, "(say %s)",message);
}
   
// Why not communicate the players positions and velocities and useful events?   
void Portugal1Info::send_game_status()   // Send Game Status
{
  char str[512];
  char str_aux[512];

  if(!MyConf()) return;

  sprintf(str_aux,"world_status %1.2f %1.2f ", MyX(),MyY());
  strcpy(str,str_aux);
  if (BallPositionValid()) {
    sprintf(str_aux,"1 %1.2f %1.2f %1.2f ",BallPositionValid(), BallX(),BallY());
    strcat(str,str_aux);
    if (BallVelocityValid())
      sprintf(str_aux,"1 %1.2f %1.2f %1.2f ",BallVelocityValid(),
                       BallAbsoluteVelocity().x, BallAbsoluteVelocity().y);
    else sprintf(str_aux,"0 ");
    strcat(str,str_aux);
  }
  else {
    sprintf(str_aux,"0 ");
    strcat(str,str_aux);
  }
  say_something(str);
}

// To compress and encode this, is a very good idea!
void Portugal1Info::parse_world_status(char *message)   // Parse Game Status
{
  char *str = &message[0];
  float x, y, pivotx, pivoty, conf, vx, vy, vconf;

  get_word(&str);
  pivotx = get_float(&str); pivoty = get_float(&str);  
  if (get_int(&str)) { // Ball position
    conf = get_float(&str); x = get_float(&str); y = get_float(&str);
    if (get_int(&str)) { // Ball Velocity
      vconf  = get_float(&str); vx = get_float(&str); vy = get_float(&str);
      HearBall(x,y,conf, vx,vy,vconf,Vector(x-pivotx,y-pivoty).mod(), CurrentTime);
    }
    else HearBall(x,y,conf, Vector(x-pivotx,y-pivoty).mod(), CurrentTime);
  }
}

// To use an online coach may be very useful!
void Portugal1Info::parse_received_message(int direction, char *message)
{
  switch(message[0]) { case 'w': parse_world_status(message); }
}


/****************************************************************************************/
/************************** Individual Decision Making  *********************************/
/****************************************************************************************/

// Lots of things to explore here! This is really a basic decision mechanism...
void Portugal1Info::DecideAction(void)   // Individual Decision Making
{
   if(!MyConf() || !BallPositionValid()) {scan_field_with_body(); return;}
   RefreshPositioningMatrix();  // analyses interceptions
   if (BallKickable()) DoBallMove();       // I have the ball? What to do with it?
   else DoPositioning();    	// Go to best Position!
}

// Pass, Forward, Dribble, Hold, Shoot to different places, [Reis and Lau 2001]...
void Portugal1Info::DoBallMove(void)   // Ball Possession: Situation With Ball
{
   smart_kick_hard(AngleToFromBody(Vector(52.5,(2*(CurrentTime.t%2)-1)*6.0)), KM_Moderate);  //Shoot to the Opponents Goal Corner!
}

// It is not a good idea to always intercept... Sometimes is better to keep the position... 
// Think about other ball recovery possibilities like mark pass lines, mark opponents, cover goal, ...
void Portugal1Info::DoPositioning(void)   // Ball Recovery and Positioning: Situations Without ball
{
  if (PO_MyInterc-2 <= PO_TeammateInterc && PO_MyInterc <= 30 || PO_MyInterc <= 10) get_ball();  // Intercept the Ball   
  else DoSBSPPositioning(); // Strategic Position
}

// This is not really SBSP [Reis and Lau 2001] (it is SP) because situations are not considered. Why 
// not convert this to SBSP? Reason about offside lines, changing of formations in some situations, ...
void Portugal1Info::DoSBSPPositioning(void)   // Strategic Situation
{
  go_to_point(SBSPPosition(MyNumber),1.5+fabs(BallX()-MyX())/10.0,PowerConserveStamina()); 
}

// Of course this is also basic. Reason about the importance of spending stamina at each moment!
float Portugal1Info::PowerConserveStamina(void)  // Stamina Management
{
  if (MyInterceptionAble() && 
     (abs(PO_MyInterc-PO_OpponInterc)<3 || PO_MyInterc<4)) return 100.0;  // if Ball is disputed 
  if(MyStamina()-150.0<SP_recover_dec_thr*SP_stamina_max) return 30.0*(MyRecovery()); // don't loose recovery!
  return 100.0;
}

// Only basic information included. Why not use interception matrixes (with values for all players)
// and reason using that information. Use distances on the reasoning! Use similar structure to
// Ball Possession information!
void Portugal1Info::RefreshPositioningMatrix()  //Interception Analysis
{
  if (FastestOpponentToBall() != Unum_Unknown) {	// Number of cycles to ball of the opponent 
     PO_OpponInterc = OpponentInterceptionNumberCycles(FastestOpponentToBall());}
  else { PO_OpponInterc = 150; }
  if (FastestTeammateToBall() != Unum_Unknown) {	// Number of cycles to ball of my fastest teammate
     PO_TeammateInterc = TeammateInterceptionNumberCycles(FastestTeammateToBall());}
  else { PO_TeammateInterc = 200; }
  if (MyInterceptionAble()) {     			// My number of cycles to ball and my interception point
     PO_MyInterc = MyInterceptionNumberCycles(); }
  else { PO_MyInterc = 250; }
}

/****************************************************************************************/
/******************** SBSP - Situation Based Strategic Positioning  *********************/
/****************************************************************************************/

// Calculates the Strategic Positioning using a very simple SBSP [Reis and Lau 2001] implementation
// Improve this using situations, attractions by points or lines in some situations, etc.
Vector Portugal1Info::SBSPPosition(int t)
{
  Vector Position;
  float FactorX, FactorY;

  Position = FormationGetPosition(t);
  FactorX = FormationGetAttraction(t).x;
  FactorY = FormationGetAttraction(t).y;

  Position.x = Position.x + BallX() * FactorX;
  Position.y = Position.y + BallY() * FactorY;
  if (fabs(Position.y) > 24.0) Position.y = Position.y - 0.5*FactorY*BallY();

  Position = PositionToOnsidePosition(Position);           
  if (FormationGetBehindBall(t) && BallX() < Position.x) Position.x = BallX();
  if (Position.x > FormationGetMaxX(t)) Position.x = FormationGetMaxX(t); 
  if (Position.x < FormationGetMinX(t)) Position.x = FormationGetMinX(t); 

  return Position;
}

/****************************************************************************************/
/*************************  Normal Player Central Module  *******************************/
/****************************************************************************************/

// This is a very simple player. Use different things like DPRE - Dynamic Positioning and Role
// Exchange ([Reis and Lau 2001] that enables the players to switch their positionings in a given
// formation or SLM - Strategic Looking Mechanism ([Reis and Lau 2001]) that enables the players to
// look to llok at the best spot in the field instead of looking at the ball!
void Portugal1Info::NormalPlayer()
{
   static int players_moved = FALSE;
   if (PlayMode == PM_Before_Kick_Off || PlayMode == PM_Half_Time || PlayMode == PM_Extended_Time)
	if (players_moved) return; 
	else {
       	      players_moved = TRUE;
      	      if (FormationGetPosition(MyNumber).x < 0.0) move(FormationGetPosition(MyNumber));
	      else move(-1.0,FormationGetPosition(MyNumber).y);
              change_view(VW_Narrow); //Use Narrow vision? Maybe only in some situations...
       	      return;
   	  }
   players_moved = FALSE;

   DecideFormation();		// Formation Analysis
   DoCommunication();		// Communicate
   DecideAction();  		// Individual Decision Making
}

/****************************************************************************************/
/*************************  Goalie Player Central Module  *******************************/
/****************************************************************************************/

// This is a very simple goalie. A decent goalie [Reis and lau 2001] must look always (using 45 
// degree vision) to the ball and must not try risky interceptions. Backward dashing is needed! 
// Try to see if the opponent has the ball controled and use this on the reasoning process!
void Portugal1Info::GoaliePlayer()
{
  static int cat_time, First_time = TRUE; 
  static Line pen_border,pen_top,pen_bottom;
  Vector Goalie_PosTop   =Vector( -SP_pitch_length/2.0+SP_penalty_area_length, -SP_penalty_area_width/2.0 ) + Vector(-5.0,10.0);
  Vector Goalie_PosBottom=Vector( -SP_pitch_length/2.0+SP_penalty_area_length, +SP_penalty_area_width/2.0 ) + Vector(-5.0,-10.0);
  Vector pos;
  Line ball_line;
       
  if (First_time) { 
       First_time = FALSE; 
       cat_time = 0;
       move(SBSPPosition(MyNumber)); //initialize
       pen_border.LineFromTwoPoints(Goalie_PosTop,Goalie_PosBottom); 
       pen_top.LineFromTwoPoints( Goalie_PosTop, Vector(-50.0,Goalie_PosTop.y));
       pen_bottom.LineFromTwoPoints(Goalie_PosBottom, Vector(-50.0,Goalie_PosBottom.y));
       change_view(VW_Narrow); //Use Narrow vision? Be careful not to loose the ball...
       return;
   }    
   cat_time++;

   if (PlayMode== PM_My_Goalie_Free_Kick && cat_time==1 ) { move(Vector(-39.0,-13.0*MyY()/fabs(MyY()))); return; } //move
   if (BallCatchable() && cat_time >= 6) { goalie_catch(BallAngleFromBody()); cat_time = 0; return; } //catch
   if (BallKickable()) { smart_kick_hard(AngleToFromBody(Vector(0.0,MyY()*2.0)), KM_Moderate); return; } //kick 
   
   if (BallPositionValid() && MyConf()) {
        if (BallInOwnPenaltyArea() || MyInterceptionAble() && InOwnPenaltyArea(MyInterceptionPoint())) {
	  if (MyInterceptionAble()) get_ball(); //intercept
	  else go_to_point(BallAbsolutePosition(),0.5); //goto ball
	}
	else {
          ball_line.LineFromTwoPoints(MarkerPosition(RM_My_Goal),BallAbsolutePosition());
          pos = pen_border.intersection(ball_line);
          if (pos.y > Goalie_PosBottom.y) pos = pen_bottom.intersection(ball_line);
	  else if (pos.y < Goalie_PosTop.y) pos = pen_top.intersection(ball_line);
          go_to_point(pos,1.0+BallDistance()/20.0); //goto mark position
        }
   }
   else { scan_field_with_body(); }
}


