/* -*- Mode: C++ -*- */

/* MemPortugal.h
 *
 * FCP Basic Agent
 * Luis Paulo Reis, Nuno Lau 2000 - Very simple version for demonstration
 * Copyright (C) 2000 Luis Paulo Reis, Nuno Lau
 *
 * Contacts:
 * Luis Paulo Reis 
 *    University Fernando Pessoa - Porto, Portugal
 *    Http://www.cerem.ufp.pt/~lpreis
 *    Email: lpreis@ufp.pt
 *    LIACC - Artificial Intelligence and Computer Science Lab.- University of Porto
 *    Http://www.fe.up.pt/~lpreis
 *    Email: lpreis@fe.up.pt
 * Nuno Lau
 *    University of Aveiro - Aveiro, Portugal
 *    Http://www.ua.pt/~lau
 *    Email: lau@det.ua.pt
 *
 * This Team is based on CMUnited99 source code from Peter Stone, Patrick Riley and 
 * Manuela Veloso. If you make any changes to CMUnited99 code, or have any comments 
 * about it, please send a message to Peter Stone (pstone@cs.cmu.edu).
 *
 * You may copy and distribute this code freely as long as you retain this notice.
 * If you think this simple agent was useful to you for creating a good team or 
 * if you want to discuss some RoboCup related issues, please send me an Email 
 * message (lpreis@fe.up.pt).
 *
 * We tried to keep everything simple, using the minimum amount of code. We also 
 * tried to create a competitve team and give ideas about how to transform it into 
 * a very good team!
 * 
 * This Team is able to play a decent game against all RoboCup99 teams (including 
 * CMUnited99) and RoboCup2000 teams (including FC Portugal). If you want to transform 
 * this simple team into a really competitive team, please read the suggestion 
 * included in the code and in the paper: [Reis and Lau 2001] Reis, Luis Paulo and 
 * Lau, Nuno, "FC Portugal Team Description: RoboCup 2000 Simulation League Champion", 
 * RoboCup 2000: Robot Soccer World Cup IV, Springer Verlag, 2001.
 *
 * For more information about the FC Portugal Project (European and World Champions of 
 * RoboCup 2000), please see Http://www.ieeta.pt/robocup or contact us.
 */

#ifndef _MEMPORTUGAL_H_
#define _MEMPORTUGAL_H_
#include "MemAction.h"

#define NUM_FORM 1      //Formations (like 433Open, 422, etc)
#define NUM_PLTYPES 5   //Player Types

class Portugal1Info : public ActionInfo{
public:
  int FM_Positioning[12];
  Vector FM_Position[NUM_FORM+1][1][12];
  int FM_PlayerType[NUM_FORM+1][12];
  Vector FM_Attraction[NUM_FORM+1][NUM_PLTYPES+1];
  int FM_BehindBall[NUM_FORM+1][NUM_PLTYPES+1];
  float FM_MinX[NUM_FORM+1][NUM_PLTYPES+1];
  float FM_MaxX[NUM_FORM+1][NUM_PLTYPES+1];

  int PO_OpponInterc;
  int PO_TeammateInterc;
  int PO_MyInterc;

  void NormalPlayer();
  void GoaliePlayer();

  void Initialize();
  void ReadFormations();
  void DecideFormation(void);

  void DoCommunication(void);
  void say_something(char *message);
  void send_game_status();
  void parse_world_status(char *message);
  void parse_received_message(int direction,char *message);

  void DecideAction(void);
  void DoBallMove(void);
  void DoPositioning(void);
  void DoSBSPPositioning(void);
  Vector SBSPPosition(int t);
  void RefreshPositioningMatrix();
  float PowerConserveStamina(void);

  inline int  GetGlobalFormation(void) { return n_globalform; }
  inline void SetGlobalFormation(int select) { n_globalform=select; }
  int GetPositioning(int t) { return FM_Positioning[t];}
  Vector FormationGetPosition(int t) { return FM_Position[n_globalform][0][GetPositioning(t)];}
  int FormationGetPlayerType(int t) { return FM_PlayerType[n_globalform][GetPositioning(t)];}
  Vector FormationGetAttraction(int t) { return FM_Attraction[n_globalform][FormationGetPlayerType(t)];}
  int FormationGetBehindBall(int t) { return FM_BehindBall[n_globalform][FormationGetPlayerType(t)];}
  float FormationGetMinX(int t) { return FM_MinX[n_globalform][FormationGetPlayerType(t)];}
  float FormationGetMaxX(int t) { return FM_MaxX[n_globalform][FormationGetPlayerType(t)];}
 
private:
  int       n_globalform;           //Formation
};

#endif

